#! /bin/bash

# Personal script for building the Linux packages
# and the source packages on PCLinuxOS/Mageia

# Last updated 6th October 2020 by Martin Foster.

# [1] Grab the OpenXP source and build the source
# tarball and the source ZIP package.
#
# [2] Install the Indy10 source and build the RPM package.
#
# [3] Build the ZIP package from the RPM package.

vernum=47
relnum=1
srcdir=openxp-5.0.$vernum.src

if [ ! -d ~/Build_ALL/$srcdir ] ; then
  mkdir -p ~/Build_ALL/$srcdir
fi

cd ~/Build_ALL/$srcdir

##################################################
# ========== Create the source archives ======== #
##################################################

# Grab the OpenXP source .....
# We don't need the svn info so we 'export' instead of 'checking out'
svn export --force https://svn.code.sf.net/p/openxp5/code ~/Build_ALL/$srcdir

# Create the source tarball EXCLUDING the Indy10 source
cd ..
tar -zcvf $srcdir.tar.gz $srcdir

# Move the source tarball ready for release
mv $srcdir.tar.gz ~/Projects/openxp/dist/src

# Create the source zipfile EXCLUDING the Indy10 source
zip -r $srcdir.zip $srcdir

# Move the source zipfile ready for release
mv $srcdir.zip ~/Projects/openxp/dist/src

#################################################
# ============ Build the RPM package ========== #
#################################################

# We don't do it this way anymore .....
# Grab the Indy10 source and create the source tarball INCLUDING the Indy10 source
#svn export --username=Indy-Public-RO --password="" --force https://svn.atozed.com:444/svn/Indy10/trunk@5499 ~/Build_ALL/$srcdir/Indy10
#

# We now do it this way .....
# Copy existing PRISTINE Indy10 source from the "Working Copy"
cp -R ~/Projects/openxp/Indy10/Pristine/Lib ~/Build_ALL/$srcdir/Indy10
#

# Prepare for patching .....
# Convert DOS line-endings to UNIX
cd $srcdir/Indy10
dos2unix Indy10.patch

cd Lib/Core
dos2unix IdCompilerDefines.inc

cd ../Protocols
dos2unix IdCompilerDefines.inc
dos2unix IdMessage.pas
dos2unix IdMessageClient.pas
dos2unix IdNNTP.pas
dos2unix IdResourceStringsOpenSSL.pas
dos2unix IdSSLOpenSSLHeaders.pas

cd ../System
dos2unix IdCompilerDefines.inc
dos2unix IdGlobal.pas

# We don't do this anymore .....
# Apply the Indy10 patch
#cd ../..
#patch -p1 -N < Indy10.patch
#

# Build the source tarball
cd ~/Build_ALL
tar -zcvf $srcdir.tar.gz $srcdir

# Prepare to build the RPM package

# PCLOS
#mv -f $srcdir.tar.gz ~/src/rpm/SOURCES
#cd $srcdir/Indy10
#cp -f Indy10.patch ~/src/rpm/SOURCES
#cd ../build
#cp -f openxp.spec ~/src/rpm/SPECS
#cd ~/src/rpm/SPECS

# Mageia
mv -f $srcdir.tar.gz ~/rpmbuild/SOURCES
cd $srcdir/Indy10
cp -f Indy10.patch ~/rpmbuild/SOURCES
cd ../build
cp -f openxp.spec ~/rpmbuild/SPECS
cd ~/rpmbuild/SPECS

# Build the RPM package .....
# We don't need an SRPM so we use the '-bb' switch
rpmbuild -bb openxp.spec

# Copy the RPM package ready for release
if [ $(uname -m) = 'x86_64' ]; then
  cd ../RPMS/x86_64
  cp openxp-5.0.$vernum-$relnum.x86_64.rpm ~/Projects/openxp/dist/linux
else
  cd ../RPMS/i586
  cp openxp-5.0.$vernum-$relnum.i586.rpm ~/Projects/openxp/dist/linux
fi

##################################################
# ========== Create the ZIP package ============ #
##################################################

# *Copy* the RPM package ready for extraction
if [ $(uname -m) = 'x86_64' ]; then
  cp openxp-5.0.$vernum-$relnum.x86_64.rpm ~/Build_ALL
else
  cp openxp-5.0.$vernum-$relnum.i586.rpm ~/Build_ALL
fi

# Extract the RPM package
cd ~/Build_ALL
if [ $(uname -m) = 'x86_64' ]; then
  rpm2cpio openxp-5.0.$vernum-$relnum.x86_64.rpm | cpio -idmv
else
  rpm2cpio openxp-5.0.$vernum-$relnum.i586.rpm | cpio -idmv
fi

# Remove the RPM package
if [ $(uname -m) = 'x86_64' ]; then
  rm openxp-5.0.$vernum-$relnum.x86_64.rpm
else
  rm openxp-5.0.$vernum-$relnum.i586.rpm
fi

# Prepare to build the ZIP package
cd usr/local/lib/openxp
rm -f bin/notice.sh
cp ~/Projects/openxp/dist/linux/install.sh install.sh
cp ~/Projects/openxp/dist/linux/readme.1st readme.1st
cp ~/Projects/openxp/dist/linux/readme.2nd readme.2nd

# Build the ZIP package
# IMPORTANT NOTES:
# The ZIP package name must NOT exceed 30 characters.
# This is a workaround for Fido file request processors
# which choke on filenames greater than 30 characters
if [ $(uname -m) = 'x86_64' ]; then
  zip -r openxp-5.0.$vernum-$relnum.x86_64-lnx.zip .
else
  zip -r openxp-5.0.$vernum-$relnum.i586-lnx.zip .
fi

# Copy the ZIP package ready for release
if [ $(uname -m) = 'x86_64' ]; then
  cp openxp-5.0.$vernum-$relnum.x86_64-lnx.zip ~/Projects/openxp/dist/linux
else
  cp openxp-5.0.$vernum-$relnum.i586-lnx.zip ~/Projects/openxp/dist/linux
fi

# Cleanup
cd
rm -fr ~/Build_ALL
