{  $Id: ncpop3.pas 7175 2005-11-12 16:14:51Z mkaemmerer $

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this software; see the file gpl.txt. If not, write to the
   Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on August, 1st 2000 by Markus Kmmerer <mk@happyarts.de>

   This software is part of the OpenXP project (www.openxp.de).
}

{ Klasse TPOP3 }

{$I xpdefine.inc}

unit ncpop3;

interface

uses
  SysUtils, Classes, NetcallIndy, IdPOP3, IdException;

type
  EPOP3                 = EIdException;       { Vorfahr }
  EPOP3Local            = class(EPOP3);       { Allgemein }
  EPOP3InvalidParameter = class(EPOP3Local);  { Ungueltiger Parameter }

type
  TMyIdPOP3 = class(TIdPOP3)
  public
    property IPVersion;
  end;

  TPOP3 = class(TNetcallIndy)
  private
    IdPOP3              : TMyIdPOP3;

    FHost               : string;       { Hostadresse }
    FPort               : integer;      { Portnummer }
    FIPv6               : boolean;
    FSSLMethod          : integer;
    FUseExplicitTLS     : boolean;

    FUseAPOP            : Boolean;   { APOP = encrypted passwords }
    FOnlyNew            : Boolean;   { nur neue Mails holen }
    FUser, FPassword    : string;    { Identifikation }
    FMailCount,FLastRead: Integer;
    FMailSize           : Int64;
    FSupportUIDLs       : Boolean;   { Flag: Server supports UIDLs }
    FAvailableUIDLs     : TStringList; { NEW available mail UIDLs }
    FMaxMailSize        : Integer;   { maximum size of mail to get }

    procedure InitVars;

  protected
    procedure SetPort(i: integer);
    procedure SetSSLMethod(i: integer);
    procedure SetUseExplicitTLS(b: boolean);

    { Return MailCount-LastReadID or FAvailableUIDLs.Count depending
      on whether LAST is implemented or not }
    function GetNewMailCount: Integer;
    function GetLastRead: Integer;
    function MapUIDL(Nr: Integer): Integer;

  public
    { List of already retrieved mail UIDLs, used and updated by
      RetrAll/LastRead/Retr }
    UIDLs: TStringList;

    constructor Create;
    constructor CreateWithHost(s: string);
    destructor Destroy; override;

    property Port:integer read FPort write SetPort;
    property IPv6:boolean read FIPv6 write FIPv6;
    property SSLMethod:integer read FSSLMethod write SetSSLMethod;
    property UseExplicitTLS:boolean read FUseExplicitTLS write SetUseExplicitTLS;

    property User: string read FUser write FUser;
    property Password: string read FPassword write FPassword;
    property UseAPOP: Boolean read FUseAPOP write FUseAPOP;
    { Take only new mails into account, remap mail numbers (if UIDL support) -
      initialize before Stat }
    property OnlyNew: Boolean read FOnlyNew write FOnlyNew;
    { Count of mails in mbox }
    property MailCount: Integer read FMailCount;
    { Count of NEW mails }
    property NewMailCount: Integer read GetNewMailCount;
    function MailSize: Cardinal;
    { Number of last read mail - 0 with UIDL support and OnlyNew}
    property LastRead: Integer read GetLastRead;
    property MaxMailSize: Integer read FMaxMailSize write FMaxMailSize;

    { Verbindung herstellen und Anmelden (Auto-Login) }
    function Connect: boolean;
    function Connected: boolean;

    { Abmelden }
    procedure Disconnect;

    { -------- POP3-Zugriffe }

    // Initializes UIDL and statistical variables, should be called after Connect
    function Stat: boolean;
    // Empfngt eine Nachricht
    function Retr(ID: Integer; List: TStringList; var InfoString: string): boolean;
    // Empfngt alle Nachrichten
    function RetrAll(List: TStringList): boolean;
    // Lscht die angegebene Nachricht
    function Dele(ID: Integer): boolean;
    // Lscht High-Watermark und als gelscht markierte Nachrichten
    function RSet: boolean;
  end;

implementation

uses
  xpglobal, typeform, ProgressOutput, ncresource,
  IdExplicitTLSClientServerBase, IdSSLOpenSSL, IdIOHandler, IdReplyPOP3, IdMessage,
  IdGlobal;

const
  DefaultPOP3Port       = 110;


constructor TPOP3.Create;
begin
  inherited Create;
  InitVars;
end;

constructor TPOP3.CreateWithHost(s: string);
begin
  Create;
  FHost := s;
end;

procedure TPOP3.InitVars;
begin
  FHost:='';
  FPort:=DefaultPOP3Port;
  FUser:='';
  FPassword:='';

  FUseAPOP := True;
  FOnlyNew := True;
  FLastRead:= -1;
  FSupportUIDLs:= False;
  UIDLs:=TStringList.Create;
  FAvailableUIDLs:=TStringList.Create;

  IdPOP3 := TMyIdPOP3.Create;
  IdPOP3.ReadTimeout := DefaultTimeout;
  IdPOP3.OnConnected := OnConnected;
end;

destructor TPOP3.Destroy;
begin
  try
    Disconnect;
  except
  end;
  IdPOP3.Free;
  UIDLs.Free;
  FAvailableUIDLs.Free;
  inherited Destroy;
end;

function TPOP3.Connect: boolean;
begin
  // Authorisierung bei POP3 immer ntig
  if (FUser='') or (FPassword='') then
    raise EPOP3Local.CreateFmt(res_connect3, ['Invalid account info']); // Anmeldung fehlgeschlagen

  IdPOP3.Host := FHost;
  IdPOP3.Port := FPort;
  if (IPv6) then
    IdPOP3.IPVersion := Id_IPv6
  else
    IdPOP3.IPVersion := Id_IPv4;

  IdPOP3.Username := FUser;
  IdPOP3.Password := FPassword;
  IdPOP3.AutoLogin := True;
  if (FUseAPOP) then
    IdPOP3.AuthType := patAPOP
  else
    IdPOP3.AuthType := patUserPass;

  Output(mcVerbose,res_connect1, [FHost]);

  IdPOP3.Connect;

  Output(mcInfo,res_connect4, [FHost]); // Verbunden

  Result:= true;
end;

procedure TPOP3.Disconnect;
begin
  if Connected then
  begin
    Output(mcInfo,res_disconnect);
    try
      IdPOP3.Disconnect;
    except
    end;
  end;
end;

function TPOP3.Connected: boolean;
begin
  Result := IdPOP3.Connected;
end;

function TPOP3.MailSize: Cardinal;
begin
  // avoid RangeCheck-Exception
  if (FMailSize <= Low(Cardinal)) then
    Result := Low(Cardinal)
  else if (FMailSize >= High(Cardinal)) then
    Result := High(Cardinal)
  else
    Result := Cardinal(FMailSize);
end;

function TPOP3.GetNewMailCount: Integer;
begin
  if FSupportUIDLs then
    result := FAvailableUIDLs.Count
  else
    result := MailCount - FLastRead;
end;

function TPOP3.GetLastRead: Integer;
begin
  Result := FLastRead;
  if OnlyNew and FSupportUIDLs then
    Result := 0;
end;

function TPOP3.Stat: Boolean;
var
  s: String;
  UIDL: String; Nr: Integer;
  ServerUIDLs: TStringList;
  i: integer;
begin
  Result := false;
  if not Connected then
    exit;

  if (not IdPOP3.RetrieveStats(FMailCount, FMailSize)) then
    Exit;

  Result := true;

  ServerUIDLs := TStringList.Create;
  try
    FSupportUIDLs := IdPOP3.UIDL(ServerUIDLs);
    if (FSupportUIDLs) then
    begin
      for i := 0 to ServerUIDLs.Count-1 do
      begin
        s := ServerUIDLs[i];
        UIDL := Mid(s, cPos(' ', s) + 1);
        Nr := StrToIntDef(LeftStr(s, cPos(' ', s) - 1), 0);
        if UIDLs.IndexOf(UIDL) = -1 then
          FAvailableUIDLs.AddObject(UIDL, Pointer(Nr));
        ServerUIDLs[i] := UIDL;
      end;
      // we're done, now delete old UIDLs not available from server anymore
      // from UIDL list to prevent it from getting big
      Nr := 0;
      while Nr < UIDLs.Count do
      begin
        if ServerUIDLs.IndexOf(UIDLs[Nr]) = -1 then
          UIDLs.Delete(Nr)
        else
          Inc(Nr);
      end;
      FLastRead := 0;
    end;
  finally
    ServerUIDLs.Free;
  end;

  if (IdPOP3.SendCmd('LAST','') <> ST_OK) then
  begin
    if (FLastRead = -1)and(FOnlyNew) then
      raise EPOP3Local.Create(res_nolastinfo)
    else
      exit;
  end;
  s := IdPOP3.LastCmdResult.Text[0];
  s := Copy(s, 5, Length(s));
  s := Copy(s, 1, cPos(' ', s) - 1);
  FLastRead := StrToIntDef(s, 0);
end;


function TPOP3.MapUIDL(Nr: Integer): Integer;
begin
  if (not FSupportUIDLs)or(not OnlyNew) then
    result:=Nr
  else
    result:=Int64(Pointer(FAvailableUIDLs.Objects[Nr - 1]));
end;

function TPOP3.Retr(ID: Integer; List: TStringList; var InfoString: string): boolean;
var
  Nr: Integer;
  MailSize: Integer;
  IdMessage: TIdMessage;
begin
  Result := False;
  InfoString := '';
  if Connected then
  begin
    ID := MapUIDL(ID);

    if MaxMailSize > 0 then
    begin
      IdPOP3.SendCmd(Format('LIST %d',[ID]),'');
      MailSize := StrToIntDef(ExtractWord(3, IdPOP3.LastCmdResult.Text[0]), 0);
    end
    else
      MailSize := 0;

    if (MailSize = 0) or (MailSize <= MaxMailSize * 1024) then
    begin
      IdMessage := TIdMessage.Create;
      try
        if (IdPOP3.Retrieve(ID, IdMessage)) then
        begin
          InfoString := IdMessageToStringList (IdMessage, List);
          Result := True;
          if FSupportUIDLs then
          begin
            // mark UIDL as retrieved
            for Nr := 0 to FAvailableUIDLs.Count - 1 do
            begin
              if Int64(Pointer(FAvailableUIDLs.Objects[Nr])) = ID then
              begin
                if UIDLs.IndexOf(FAvailableUIDLs[Nr]) = -1 then
                  UIDLs.Add(FAvailableUIDLs[Nr]);
                break;
              end;
            end;
          end;
        end
        else
          InfoString := res_msg_not_avail + #13#10;
      finally
        IdMessage.Free;
      end;
    end;
  end;
end;

function TPOP3.Dele(ID: Integer): boolean;
begin
  Result := False;
  if Connected then
  begin
    if (IdPOP3.SendCmd(Format('DELE %d', [MapUIDL(ID)]),'') = ST_OK) then
      Result := True;
  end;
end;

function TPOP3.RetrAll(List: TStringList): boolean;
var
  i: Integer;
  FirstMail: Integer;
  InfoStringDummy: string;
begin
  result:=true;
  if OnlyNew then
    FirstMail := LastRead + 1
  else
    FirstMail := 1;
  for i := FirstMail to FirstMail+NewMailCount-1 do
    result:=result and Retr(i, List, InfoStringDummy)
end;

function TPOP3.RSet: boolean;
begin
  Result := False;
  if Connected then
  begin
    if (IdPOP3.SendCmd('RSET','') = ST_OK) then
      Result := True;
  end;
end;

procedure TPOP3.SetPort(i: integer);
begin
  if (i >= 0) and (i <= 65535) then
    FPort:= i
  else
    raise EPOP3InvalidParameter.CreateFmt('Port must be in 0..65535 (%d)', [i]);
end;

procedure TPOP3.SetSSLMethod(i: integer);
var
  IOHandler: TIdIOHandler;
const
  SSLMethods: array[1..2] of TIdSSLVersion = (sslvSSLv3, sslvTLSv1_2);
begin
  IOHandler := IdPOP3.IOHandler;
  IdPOP3.IOHandler := nil;
  try
    case i of
      0: begin end;
      1,2:
        begin
          IdPOP3.IOHandler := TIdSSLIOHandlerSocketOpenSSL.Create(IdPOP3);
          TIdSSLIOHandlerSocketOpenSSL(IdPOP3.IOHandler).SSLOptions.Method := SSLMethods[i];
        end;
      else
        raise EPOP3InvalidParameter.CreateFmt('SSL-Method must be in 0=none/1=SSL/2=TLS (%d)', [i]);
    end;
  finally
    IOHandler.Free;
  end;
  SetUseExplicitTLS (FUseExplicitTLS);
end;

procedure TPOP3.SetUseExplicitTLS(b: boolean);
begin
  FUseExplicitTLS := b;
  if (IdPOP3.IOHandler <> nil) then
  begin
    if (FUseExplicitTLS) then
      IdPOP3.UseTLS := utUseExplicitTLS
    else
      IdPOP3.UseTLS := utUseImplicitTLS;
  end
  else
    IdPOP3.UseTLS := utNoTLSSupport;
end;

end.
