{
   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this software; see the file gpl.txt. If not, write to the
   Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   This software is part of the OpenXP project (www.openxp.de).
}

{ Klasse TNetcallIndy }

{$I xpdefine.inc}

unit netcallindy;

interface

uses
  SysUtils, Classes, IdMessage, Netcall;

type
  TNetcallIndy = class(TNetcall)
  protected
    function DefaultTimeout: integer;
    procedure OnConnected(Sender: TObject);
  public
    function IdMessageToStringList (IdMessage: TIdMessage; List: TStringList = nil): string;
    procedure MessageStringsToStream (Msg: TStringList; MessageStream: TStringStream);
  end;

implementation

uses
  Xp0, IdGlobal, IdTCPConnection, IdStackConsts;

function TNetcallIndy.DefaultTimeout: integer;
begin
  Result := ParTcpTimeout * 1000;
end;

procedure TNetcallIndy.OnConnected(Sender: TObject);
begin
{$IFDEF Win32 }
  with TIdTCPConnection(Sender).Socket.Binding do
  begin
    SetSockOpt(Id_SOL_SOCKET, Id_SO_RCVTIMEO, DefaultTimeout);
    SetSockOpt(Id_SOL_SOCKET, Id_SO_SNDTIMEO, DefaultTimeout);
  end;
{$ENDIF }
end;

function TNetcallIndy.IdMessageToStringList (
  IdMessage: TIdMessage; List: TStringList
): string;
var
  MailStream: TStringStream;
  NameValueSeparatorTmp: Char;
  s: string;
  i: integer;
begin
  if (List <> nil) then
  begin
    MailStream := TStringStream.Create('');
    NameValueSeparatorTmp := List.NameValueSeparator;
    try
      IdMessage.SaveToStream(MailStream);
      MailStream.Position := 0;
      List.LoadFromStream(MailStream);
      if (List.Count > 0) and (List[List.Count-1] = '.') then
        List.Delete(List.Count-1);
      for i := 0 to List.Count-1 do
      begin
        s := List[i];
        if (Length(s) >= 2) and (s[1] = '.') and (s[2] = '.') then
          List[i] := Copy(s,2,High(Integer));
      end;
    finally
      List.NameValueSeparator := NameValueSeparatorTmp;
      MailStream.Free;
    end;
  end;
  Result :=
    'From   : ' + IdMessage.From.Text + #13#10 +
    'To     : ' + IdMessage.Recipients.EMailAddresses + #13#10 +
    'Subject: ' + IdMessage.Subject + #13#10
  ;
end;

procedure TNetcallIndy.MessageStringsToStream (Msg: TStringList; MessageStream: TStringStream);
var
  i: integer;
begin
  for i := 0 to Msg.Count-1 do
  begin
    if (Msg[i] = '.') then
      MessageStream.WriteString('..')
    else
      MessageStream.WriteString(Msg[i]);
    MessageStream.WriteString(EOL);
  end;
end;

end.
