{  $Id: xpncimap.pas 6778 2004-01-17 16:33:53Z mk $

   OpenXP IMAP netcall unit
   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this software; see the file gpl.txt. If not, write to the
   Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on April, 26st 2003 by Markus Kaemmerer <mk@happyarts.de>

   This software is part of the OpenXP project (www.openxp.de).
}

{$I xpdefine.inc}

{ OpenXP IMAP netcall unit }
unit xpncimap;

interface

uses
  XPGlobal,
  XP0,
  SysUtils,
  Classes;

function GetIMAPMails(
  BoxName: string; bp: BoxPtr; Domain: String; IncomingFiles, DeleteSpoolFiles: TStringList; NetcallLogFile: string
): boolean;

implementation  { ------------------------------------------------- }

uses
  Netcall,ncsocket,ncsmtp,ncimap,ncresource,
  progressoutput,xpprogressoutputwindow,
{$ifdef NCRT}
  XPCurses,
{$endif}
  typeform,
  InOut,
  xp1,                          { dialoge }
  xp1o,
  xpnetcall,
  zcrfc,
  xp3o;                         { ForceRecipient }


function GetIMAPMails(
  BoxName: string; bp: BoxPtr; Domain: String; IncomingFiles,DeleteSpoolFiles: TStringList; NetcallLogFile: string
): boolean;
var
  List          : TStringList;
  aFile         : string;
  i             : integer;              { -----"------- }
  IMAP          : TIMAP;                { Socket }

  procedure ProcessIncomingFiles(IncomingFiles: TStringList);
  var iFile: Integer; uu: TUUZ;
  begin
    uu := TUUZ.Create;
    for iFile:=0 to IncomingFiles.Count-1 do
    try
      uu.source := IncomingFiles[iFile];
      uu.dest := ChangeFileExt(IncomingFiles[iFile], '.z');
      IncomingFiles[iFile] := uu.dest;
      uu.OwnSite := boxpar^.pointname+domain;
      uu.NoCharsetRecode := Enable_UTF8 or not (boxpar^.UUZCharsetRecode);
//    uu.ClearSourceFiles := true;
      uu.utoz;
    except
      on E: Exception do
        IMAP.Output(mcError, E.Message);
    end;
    DeleteSpoolFiles.AddStrings(uu.DeleteFiles);
    uu.free;
  end;

  procedure SaveMail;
  begin
    aFile:=OwnPath + XFerDir + IntToStr(i) + '.mail';
    try
      List.SaveToFile(aFile);
      IncomingFiles.Add(aFile);
    except
      on E: Exception do
        IMAP.Output(mcError, E.Message);
    end;
    List.Clear;
  end;

var
  POWindow      : TProgressOutputWindow;{ ProgressOutput }
  FirstMail,LastMail : Integer;
  UIDLFileName  : String;
  InfoString    : string;
begin
  Result := true;
  if bp^.IMAP_ip='' then exit; // exit immediately if no server specified
  { POWindow erstellen }
  POWindow:= TProgressOutputWindow.CreateWithSize(60,10,Format(res_IMAPinit,[BoxName]),True,NetcallLogFile);
  { Host und ... }
  IMAP:= TIMAP.CreateWithHost(bp^.IMAP_ip);
  List := TStringList.Create;
  try
    IMAP.OnlyNew := BoxPar^.IMAP_Onlynew;
    IMAP.LeaveUnchanged := (BoxPar^.IMAP_Clear <> '') and (BoxPar^.IMAP_Clear[1] = 'U');
    IMAP.DeleteMail := (BoxPar^.IMAP_Clear <> '') and (BoxPar^.IMAP_Clear[1] = 'J');
    { POWindow erstellen }
    IMAP.ProgressOutput:= POWindow;
    { Zugangsdaten uebernehmen }
    IMAP.User:= bp^.IMAP_id;
    IMAP.Password:= bp^.IMAP_pwd;
    IMAP.Port := Bp^.IMAP_Port;
    IMAP.IPv6 := iifb(ParIPv6 in [0,1], Boolean(ParIPv6), bp^.imap_IPv6);
    if (bp^.imap_SSLMethod = 'SSL')      then IMAP.SSLMethod := 1
    else if (bp^.imap_SSLMethod = 'TLS') then IMAP.SSLMethod := 2
    else                                      IMAP.SSLMethod := 0;
    IMAP.UseExplicitTLS := bp^.imap_UseExplicitTLS;

    { Get last retrieved UIDLs from file }
    UIDLFileName:=FileUpperCase(OwnPath+GetServerFilename(Boxname, extUdl));
    if FileExists(UIDLFileName)then
      IMAP.UIDLs.LoadFromFile(UIDLFileName);

    { Verbinden }
    IMAP.Connect;
    IMAP.Stat;

    IMAP.Output (mcInfo, res_mailstat_imap, [IMAP.MailCount, IMAP.NewMailCount]);

    LastMail := IMAP.MailCount;
    FirstMail := IMAP.FirstUnseen;

    for i := FirstMail to LastMail do
    begin
      if (i = FirstMail) then
        IMAP.Output(mcDefault,' ');
      IMAP.Output(mcVerbose,res_getmail,[i]);
      if (IMAP.Retr(i, List, InfoString)) then
        SaveMail;
      IMAP.Output(mcDefault,InfoString);
    end;

    IMAP.Expunge;
    IMAP.Disconnect;
  except
    on E: EIMAP do begin
      IMAP.Output(mcError, E.Message);
      result:= false;
      end;
    on E: EUserBreakError do begin
      IMAP.Output(mcError, res_userbreak);
      result:= false;
      end;
    on E: ESocketNetcall do begin
      IMAP.Output(mcError, res_noconnect);
      result:= false;
      end;
    on E: Exception do begin
      IMAP.Output(mcError, res_strange + E.Message);
      result:= false;
      end;
  end;

  if IMAP.UIDLs.Count>0 then
    IMAP.UIDLs.SaveToFile(UIDLFileName)
  else
    DeleteFile(UIDLFileName);

  List.Free;

  ProcessIncomingFiles(IncomingFiles);
  IMAP.Free;

  if BoxPar^.IMAP_ForceOneArea then begin
    // tell xp3o.PufferEinlesen to put all messages to standard mail area
    xp3o.ForceRecipient:= '1/' + BoxPar^.username + iifs(userboxname,'/'+BoxPar^.Boxname,'');
    i:= cPos('@',xp3o.ForceRecipient);
    if i>0 then
      xp3o.ForceRecipient:= LeftStr(xp3o.ForceRecipient, i - 1);
    end;
end;


{
  $Log: xpncimap.pas,v $
  Revision 1.5  2004/01/17 16:33:52  mk
  - split xp0.pas in xp0.pas and xpconst.pas to remove some dependencies
    xpconst.pas should be used for global constants (only!)

  Revision 1.4  2003/08/29 19:35:54  mk
  - if now message is available then do not try to get one

  Revision 1.3  2003/08/11 21:28:06  mk
  - fixed IMAP if OnlyNew is switched on and no new mail is waiting

  Revision 1.2  2003/05/02 07:29:41  mk
  - lowercase ncimap (for linux)

  Revision 1.1  2003/05/01 09:52:30  mk
  - added IMAP support

}
end.

