{  $Id: xpncpop3.pas 7094 2005-10-20 13:22:50Z mkaemmerer $

   OpenXP POP3 netcall unit
   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this software; see the file gpl.txt. If not, write to the
   Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on December, 26st 2000 by Markus Kaemmerer <mk@happyarts.de>

   This software is part of the OpenXP project (www.openxp.de).
}

{$I xpdefine.inc}

{ OpenXP POP3 netcall unit }
unit xpncpop3;

interface

uses
  XPGlobal,
  XP0,
  SysUtils,
  Classes;

function GetPOP3Mails(
  BoxName: string; bp: BoxPtr; Domain: String;
  IncomingFiles, DeleteSpoolFiles: TStringList; NetcallLogfile: string
): boolean;
function SendSMTPMails(BoxName,boxfile: string; bp: BoxPtr; EMail, Realname, PPFile: String; NetcallLogFile: string): boolean;

implementation  { ------------------------------------------------- }

uses
  Netcall,NCSocket,ncsmtp,ncpop3,ncresource,
  progressoutput,xpprogressoutputwindow,
  typeform,
  InOut,
  xp1,                          { dialoge }
  xp1o,
  xpnetcall,
  zcrfc,
  xp3o;                         { ForceRecipient }


function SendSMTPMails(BoxName,boxfile: string; bp: BoxPtr; EMail, Realname, PPFile: String; NetcallLogFile: string): boolean;

  const RFCFile= 'SMTPTEMP';

  procedure ZtoRFC(boxpar: boxptr; source: String; const Dest: string);
  var uu: TUUZ;
  begin
    MakeMimetypCfg;
    with boxpar^ do begin
      uu := TUUZ.Create;
      uu.SMTP := true;
      uu.PPP := true;
      if MIMEqp then uu.MakeQP := true;
      if RFC1522 then uu.RFC1522 := true;
      uu.MailUser := BoxPar^.UserName;
      uu.NewsUser := BoxPar^.UserName;
      uu.FileUser := BoxPar^.UserName;
      OutFilter(source);
//    uu.ClearSourceFiles := false;
      uu.Source := source;
      uu.Dest := dest;
      uu._from := boxpar^.pointname;
      uu._to := boxpar^.boxname;
      uu.ztou;
      uu.Free;
    end;
  end;

var
  SMTP          : TSMTP;                { Socket }
  POWindow      : TProgressOutputWindow;{ ProgressOutput }
  List          : TStringList;
begin
  result:= true;
  if bp^.smtp_ip='' then exit; // exit immediately if no server specified
  DeleteFile(RFCFile);
  ZtoRFC(bp,PPFile,RFCFile);
  if not FileExists(RFCFile) then exit;

  { ProgressOutput erstellen }
  POWindow:= TProgressOutputWindow.CreateWithSize(60,10,Format(res_smtpinit,[BoxName]),True,NetcallLogFile);
  { Host und ... }
  SMTP:= TSMTP.CreateWithHost(bp^.smtp_ip);
  List := TStringList.Create;
  try
    { POWindow erstellen }
    SMTP.ProgressOutput:= POWindow;
    { Zugangsdaten uebernehmen }
    SMTP.User:= bp^.smtp_id;
    SMTP.Password:= bp^.smtp_pwd;
    SMTP.SecureLoginMandatory := bp^.smtp_secureloginmandatory;
    SMTP.Port := bp^.Smtp_Port;
    SMTP.IPv6 := iifb(ParIPv6 in [0,1], Boolean(ParIPv6), bp^.smtp_IPv6);
    if (bp^.Smtp_SSLMethod = 'SSL')      then SMTP.SSLMethod := 1
    else if (bp^.Smtp_SSLMethod = 'TLS') then SMTP.SSLMethod := 2
    else                                      SMTP.SSLMethod := 0;
    SMTP.UseExplicitTLS := bp^.Smtp_UseExplicitTLS;

    List.LoadFromFile(RFCFile);
    SMTP.Connect(SMTP.GetFQDomain(List));
    SMTP.PostPlainRFCMails(List, EMail, Realname);
    SMTP.Disconnect;
  except
    on E: ESMTP do begin
      SMTP.Output(mcError, E.Message);
      result:= false;
      end;
    on E: EUserBreakError do begin
      SMTP.Output(mcError, res_userbreak);
      result:= false;
      end;
    on E: ESocketNetcall do begin
      SMTP.Output(mcError, res_noconnect);
      result:= false;
      end;
    on E: Exception do begin
      SMTP.Output(mcError, res_strange + E.Message);
      result:= false;
      end;
  end;
  List.Free;
  SMTP.Free;

  if result then begin
    ClearUnversandt(PPFile,BoxName, nil);
    SafeDeleteFile(PPFile);
    SafeDeleteFile(RFCFile);
  end;
end;


function GetPOP3Mails(
  BoxName: string; bp: BoxPtr; Domain: String;
  IncomingFiles, DeleteSpoolFiles: TStringList; NetcallLogfile: string
): boolean;
var
  List          : TStringList;
  aFile         : string;
  i             : integer;              { -----"------- }
  POP           : TPOP3;                { Socket }

  procedure ProcessIncomingFiles(IncomingFiles: TStringList);
  var iFile: Integer; uu: TUUZ;
  begin
    uu := TUUZ.Create;
    for iFile:=0 to IncomingFiles.Count-1 do
    try
      uu.source := IncomingFiles[iFile];
      uu.dest := ChangeFileExt(IncomingFiles[iFile], '.z');
      IncomingFiles[iFile] := uu.dest;
      uu.OwnSite := bp^.pointname+domain;
      uu.NoCharsetRecode := Enable_UTF8 or not (bp^.UUZCharsetRecode);
      // uu.ClearSourceFiles := true;
      uu.utoz;
    except
      on E: Exception do
        POP.Output(mcError, E.Message);
    end;
    DeleteSpoolFiles.AddStrings(uu.DeleteFiles);
    uu.free;
  end;

  procedure SaveMail;
  begin
    aFile:=OwnPath + XFerDir + IntToStr(i) + '.mail';
    try
      List.SaveToFile(aFile);
      IncomingFiles.Add(aFile);
    except
      on E: Exception do
        POP.Output(mcError, E.Message);
    end;
    List.Clear;
  end;

var
  POWindow      : TProgressOutputWindow;{ ProgressOutput }
  FirstMail,LastMail : Integer;
  UIDLFileName  : String;
  InfoString    : string;
begin
  Result := true;
  if bp^.pop3_ip='' then exit; // exit immediately if no server specified
  { POWindow erstellen }
  POWindow:= TProgressOutputWindow.CreateWithSize(60,10,Format(res_pop3init,[BoxName]),True,NetcallLogfile);
  { Host und ... }
  POP:= TPOP3.CreateWithHost(bp^.pop3_ip);
  List := TStringList.Create;
  try
    { POWindow erstellen }
    POP.ProgressOutput:= POWindow;
    { Daten uebernehmen }
    Pop.Port := bp^.pop3_port;
    Pop.IPv6 := iifb(ParIPv6 in [0,1], Boolean(ParIPv6), bp^.pop3_IPv6);
    if (bp^.pop3_SSLMethod = 'SSL')      then Pop.SSLMethod := 1
    else if (bp^.pop3_SSLMethod = 'TLS') then Pop.SSLMethod := 2
    else                                      Pop.SSLMethod := 0;
    Pop.UseExplicitTLS := bp^.Pop3_UseExplicitTLS;
    Pop.UseAPOP := bp^.Pop3_APOP;
    Pop.OnlyNew := bp^.Pop3_Onlynew;
    { Zugangsdaten uebernehmen }
    POP.User:= bp^.pop3_id;
    POP.Password:= bp^.pop3_pwd;
    POP.MaxMailSize := Bp^.POP3_MaxMailSize;

    { Get last retrieved UIDLs from file }
    UIDLFileName:=FileUpperCase(OwnPath+GetServerFilename(Boxname, extUdl));
    if FileExists(UIDLFileName)then
      POP.UIDLs.LoadFromFile(UIDLFileName);

    pop.connect;
    pop.stat;

    POP.Output(mcinfo, res_mailstat, [pop.mailcount, pop.newmailcount, pop.mailsize]);

    firstmail := 1; lastmail := pop.mailcount;
    if pop.onlynew then begin
      firstmail := pop.lastread + 1;
      lastmail := pop.newmailcount + firstmail - 1;
      end;

    for i := FirstMail to LastMail do
    begin
      if (i = FirstMail) then
        POP.Output(mcDefault,' ');
      POP.Output(mcVerbose,res_getmail,[i]);
      if POP.Retr(i, List, InfoString) and (List.Count = 0) then
        POP.Output(mcError,res_mailtolarge,[i])
      else
        POP.Output(mcDefault,InfoString);
      if bp^.Pop3_Clear then
        POP.Dele(I);
      SaveMail;
    end;
    POP.Disconnect;
  except
    on E: EPOP3 do begin
      POP.Output(mcError, E.Message);
      result:= false;
      end;
    on E: EUserBreakError do begin
      POP.Output(mcError, res_userbreak);
      result:= false;
      end;
    on E: ESocketNetcall do begin
      POP.Output(mcError, res_noconnect);
      result:= false;
      end;
    on E: Exception do begin
      POP.Output(mcError, res_strange + E.Message);
      result:= false;
      end;
  end;

  if POP.UIDLs.Count>0 then
    POP.UIDLs.SaveToFile(UIDLFileName)
  else
    DeleteFile(UIDLFileName);

  List.Free;

  ProcessIncomingFiles(IncomingFiles);
  POP.Free;

  if bp^.POP3_ForceOneArea then begin
    // tell xp3o.PufferEinlesen to put all messages to standard mail area
    xp3o.ForceRecipient:= '1/' + bp^.username + iifs(userboxname,'/'+bp^.Boxname,'');
    i:= cPos('@',xp3o.ForceRecipient);
    if i>0 then
      xp3o.ForceRecipient:= LeftStr(xp3o.ForceRecipient, i - 1);
    end;
end;


end.
