{ $Id: viewer.pas 7226 2006-02-19 21:26:46Z hjtaenzer $

    OpenXP Library: TMimeViewer class
    Copyright (C) 2000, Markus Kaemmerer <mk@happyarts.de>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

}

{$I xpdefine.inc}

unit viewer;

interface

uses
  xpglobal, sysutils;



type
  // Viewer for one specific message type
  TMessageViewer = class
  private
    fProg: String;              // selected viewer
    fExt: String;               // file extension for selected viewer
    fInternal: Boolean;         // unconditionally use internal viewer
    function GetInternal: Boolean;
  public
    constructor Create;
    procedure GetFromMimeType(const MimeType: String);
    procedure GetFromExtension(const Extension: String);
    procedure ViewFile(const Filename: String; Fileattach:boolean);
    property Prog: String read fProg write fProg;
    property Ext: String read fExt write fExt;
    property IsInternal: Boolean read GetInternal write fInternal;
  end;

implementation

uses
  Classes,
{$IFDEF Win32}
  Windows,
{$ENDIF}
  database, xp0, xp1, xp1o, fileio, typeform;

type
  TDelayedDelete = class (TThread)
  private
    FFileName: string;
    FDelayInSeconds: integer;
  protected
    constructor Create (const FileName: string; DelayInSeconds: integer);
    procedure Execute; override;
  public
    class procedure DeleteFile(const FileName: string; DelayInSeconds: integer = 10);
  end;

constructor TDelayedDelete.Create (const FileName: string; DelayInSeconds: integer);
begin
  inherited Create (True);
  FFileName := FileName;
  FDelayInSeconds := DelayInSeconds;
  Priority := tpLowest;
  FreeOnTerminate := True;
  Resume;
end;

procedure TDelayedDelete.Execute;
begin
  if (FDelayInSeconds > 0) then
    SysUtils.Sleep(FDelayInSeconds * 1000);
  SysUtils.DeleteFile(FFileName);
end;

class procedure TDelayedDelete.DeleteFile (const FileName: string; DelayInSeconds: integer);
begin
  TDelayedDelete.Create (FileName, DelayInSeconds);
end;


constructor TMessageViewer.Create;
begin
  inherited Create;
  IsInternal := False;
  Prog := '';
  Ext := '';
end;

procedure TMessageViewer.GetFromMimeType(const MimeType: String);

  function SeekMimeType(const MimeType: string): boolean;
  begin
    dbSeek(mimebase, mtiTyp, UpperCase(MimeType));
    SeekMimeType :=not dbBOF(mimebase) and not dbEOF(mimebase) and
              (UpperCase(MimeType) = UpperCase(dbReadStr(mimebase,'typ')));
  end;

begin
  if MimeType = '' then exit;

  // is there a match for MimeTyp?
  if SeekMimeType(MimeType) then
  begin
    Prog := dbReadNStr(mimebase,mimeb_programm);
    Ext := '.' + dbReadNStr(mimebase,mimeb_extension);
  end else
    if SeekMimeType(LeftStr(MimeType, cposx('/',MimeType))+'*') then
    begin
      Prog := dbReadNStr(mimebase,mimeb_programm);
      Ext := '.' + dbReadNStr(mimebase,mimeb_extension);
    end;
end;

procedure TMessageViewer.GetFromExtension(const Extension: String);
begin
  { HJT 19.02.2006 Start, fuer nachfolgende Aufrufe wird }
  { sonst der vormalige Viewer aufgerufen                }
  Prog := '';
  Ext  := '';
  { HJT 19.02.2006 End }
  if Extension = '' then exit;
  dbSeek(mimebase,mtiExt,UpperCase(mid(Extension,2)));
  if dbFound then
  begin
    Prog := dbReadNStr(mimebase,mimeb_programm);
    Ext := Extension;
  end;
end;

function TMessageViewer.GetInternal: Boolean;
begin
  Result := (fInternal = True) or (Prog = '');
end;

procedure TMessageViewer.ViewFile(const Filename: String; Fileattach: Boolean);
var p         : Integer;
    extprog   : string;
    orgfn     : string;
    fn1       : string;
    parfn     : string;
    Dir       : string;
    Name      : string;
begin
  fn1:='';
  orgfn:= Filename;

//  orgfn:=iifs(viewer.fn<>'',GetFileDir(fn)+GetFileName(viewer.fn),'');
  if (not ValidFileName(orgfn) or FileExists(orgfn)) and (Ext<>'') and
     (cpos('.',filename)>0) then
    orgfn:= ChangeFileExt(filename, Ext);


  if not fileattach then
  begin
  if stricmp(filename,orgfn) or not ValidFileName(orgfn) or (cpos(' ',orgfn)>0)
    then orgfn:=TempS(_filesize(filename)+5000);
    if copyfile(Filename,orgfn) then fn1:=orgfn;
    end;

  extprog:=Prog;
  orgfn:=iifs(fn1<>'',fn1,Filename);

  // Tempdatei bei aktivem DELVTMP nach TMP-????.??? umbenennen
  if not fileattach and delviewtmp then
  Begin
    parfn:=TempS(_filesize(Filename)+5000);
    parfn:=LeftStr(parfn,length(parfn)-8)+'TMP-'+RightStr(parfn,8);
    end
  else parfn:=orgfn;

  // Korrekte File-extension verwenden
  ParFN := ChangeFileExt(ParFN, ExtractFileExt(Orgfn));
  if (RenameFile(orgfn,parfn) = false) then
  begin
{$IFDEF Win32}
    // Unter Windows schlgt 'RenameFile' fehl wenn die Zieldatei schon existiert, daher WinAPI verwenden
    MoveFileEx(PChar(orgfn),PChar(parfn),MOVEFILE_COPY_ALLOWED or MOVEFILE_REPLACE_EXISTING);
{$ENDIF}
  end;

  p:=pos('$FILE',UpperCase(extprog));
  if p=0 then extprog:=extprog+' '+parfn
  else extprog:=LeftStr(extprog,p-1)+parfn+mid(extprog,p+5);
(*  urep(extprog,'$TYPE',viewer.typ);
  urep(extprog,'$EXT',viewer.ext); *)
  if not XPWinShell(extprog,parfn,600,1,fileattach) then
    if not fileattach and (fn1<>'') then
      TDelayedDelete.DeleteFile(parfn,60);
end;

end.
