{   $Id: xp_des.pas 7154 2005-11-05 13:43:04Z mkaemmerer $

    OpenXP DES unit

    Copyright (C) 1991-2001 Peter Mandrella
    Copyright (C) 2000-2002 OpenXP team (www.openxp.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}

{$I xpdefine.inc }

{ OpenXP DES unit }
unit xp_des;

interface

uses
{$IFDEF NCRT }
  xpcurses,
{$ENDIF }
  inout,maus2,xp0;

function DES_PW(keystr:string): TObject;
function BLOWFISH_PW(keystr:string): TObject;
procedure BlockCipher_code_ecb(Cipher: TObject; decode:boolean; var data; ofs,total:longint; size:integer; x,y:byte);
procedure StreamCipher_code_cfb(Cipher: TObject; decode:boolean; var data; ofs,total:longint; size:integer; x,y:byte; IV:string);

implementation  {--------------------------------------------}

uses
  winxp, BlockCiphers;

type
     keytyp = array[1..8] of byte;
     sts    = array[1..8] of byte;
     stsa   = array[0..0] of sts;
     ba     = array[0..0] of byte;

// Avoid runtime errors even if compiled with R+,Q+
{$IFOPT R+}
  {$DEFINE OPT_RANGECHECK}
  {$R-}
{$ENDIF}
{$IFOPT Q+}
  {$DEFINE OPT_OVERFLOWCHECK}
  {$Q-}
{$ENDIF}

function DES_PW(keystr:string): TObject;
var
  i: integer;
  key: keytyp;
begin
  fillchar(key,sizeof(key),0);
  for i := 0 to length(keystr)-1 do
    key[i mod 8+1] := key[i mod 8+1] + ord(keystr[i+1]);
  Result := TDESCipher.Create(key, sizeof(key));
end;

function BLOWFISH_PW(keystr:string): TObject;
begin
  Result := TBlowfishCipher.Create(keystr[1], Length(keystr));
end;

procedure BlockCipher_code_ecb(Cipher: TObject; decode:boolean; var data; ofs,total:longint; size:integer; x,y:byte);
var i,j,n : integer;
    nn,tt : longint;
    p,pn  : integer;
begin
  if (Cipher <> nil) and (Cipher is TBlockCipher) then
  begin
    n:=size div 8;
    nn:=ofs div 8;
    tt:=total div 8;
    p:=100;
    for i:=0 to n-1 do
    begin
      pn:=(nn+i+1)*100 div tt;
      if (x>0) and (pn<>p) then
      begin
        gotoxy(x,y); attrtxt(col.colmboxhigh);
        moff; write(pn:3); mon;
        p:=pn;
      end;
      if decode then
        TBlockCipher(Cipher).DecryptBlock(stsa(data)[i])
      else
        TBlockCipher(Cipher).EncryptBlock(stsa(data)[i]);
    end;
    { den Rest XORen... }
    j:=size-8*n;
    for i := size-j to size-1 do
      ba(data)[i] := ba(data)[i] xor $6d;
  end;
end;

threadvar
    LastIVInt64: Int64;

procedure StreamCipher_code_cfb(Cipher: TObject; decode:boolean; var data; ofs,total:longint; size:integer; x,y:byte; IV:string);
var i,j,n : integer;
    nn,tt : longint;
    p,pn  : integer;
    IVInt64,IVInt64Save: Int64;
    TextPtr: PInt64;
    TmpSts: sts;
const
  // Konstante (8 Byte) mit bunt gemischten Bits und Bytes (dezimal: 9999999999999999999)
  NineteenNines: string = #$8A#$C7#$23#$04#$89#$E7#$FF#$FF;
begin
  if (Cipher <> nil) and (Cipher is TBlockCipher) then
  begin
    n:=size div 8;
    nn:=ofs div 8;
    tt:=total div 8;
    p:=100;
    if (ofs = 0) then
    begin
      if (IV <> '') then
      begin
        i := 1;
        while (i <= 8) and (i <= Length(IV)) and (IV[i] in ['a'..'z','A'..'Z','0'..'9',' ','-']) do
          Inc(i);
        if (i<=8) then
          IV := Copy(IV,1,i-1) + NineteenNines;
        Move (IV[1], IVInt64, sizeof(IVInt64));
      end
      else
        Move (NineteenNines[1], IVInt64, sizeof(IVInt64));
    end
    else
      IVInt64 := LastIVInt64;
    for i:=0 to n-1 do
    begin
      pn:=(nn+i+1)*100 div tt;
      if (x>0) and (pn<>p) then
      begin
        gotoxy(x,y); attrtxt(col.colmboxhigh);
        moff; write(pn:3); mon;
        p:=pn;
      end;
      TextPtr := @(stsa(data)[i]);
      TBlockCipher(Cipher).EncryptBlock(IVInt64);
      if decode then
      begin
        IVInt64Save := IVInt64;
        IVInt64 := TextPtr^;
        TextPtr^ := IVInt64 xor IVInt64Save;
      end
      else
      begin
        IVInt64 := IVInt64 xor TextPtr^;
        TextPtr^ := IVInt64;
      end;
    end;
    LastIVInt64 := IVInt64;
    j:=size-8*n;
    if (j <> 0) then
    begin
      for i := 1 to j do
        TmpSts[i] := stsa(data)[n][i];
      TextPtr := @TmpSts;
      TBlockCipher(Cipher).EncryptBlock(IVInt64);
      TextPtr^ := TextPtr^ xor IVInt64;
      for i := 1 to j do
        stsa(data)[n][i] := TmpSts[i];
    end;
  end;
end;

{$IFDEF OPT_RANGECHECK}
  {$UNDEF OPT_RANGECHECK}
  {$R+}
{$ENDIF}
{$IFDEF OPT_OVERFLOWCHECK}
  {$UNDEF OPT_OVERFLOWCHECK}
  {$Q+}
{$ENDIF}

end.
