{   $Id: xpcrt.pas 6974 2005-08-20 22:07:37Z stell $

    Copyright (C) 1997 Balazs Scheidler (bazsi@balabit.hu)
    Copyright (C) 1999 by Florian Klaempfl
    Copyright (C) 2001 OpenXP team (www.openxp.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Keyboard handling - based on Freepascal RTL
}


{$I xpdefine.inc }

unit xpcrt;

interface

uses
  Windows, XPGlobal;

function keypressed: boolean;
function readkey: char;

var
  ShiftKeyState: boolean;
  CtrlKeyState: boolean;
  StdInputHandle: THandle;

implementation

uses
  Debug, SysUtils, Mouse;

//function SetConsoleCP(cp:Windows.UINT):Windows.Bool; external 'kernel32.dll';
//function SetConsoleOutputCP(cp:Windows.UINT):Windows.Bool; external 'kernel32.dll';

var
  ScanCode : char;
  SpecialKey : boolean;
  DoingNumChars: Boolean;
  DoingNumCode: Byte;

Function RemapScanCode (ScanCode, CtrlKeyState: Word; keycode: DWord): byte;
  { Several remappings of scancodes are necessary to comply with what
    we get with MSDOS. Special Windows keys, as Alt-Tab, Ctrl-Esc etc.
    are excluded }
var
  AltKey, CtrlKey, ShiftKey: boolean;
const
  {
    Keypad key scancodes:

      Ctrl Norm

      $77  $47 - Home
      $8D  $48 - Up arrow
      $84  $49 - PgUp
      $8E  $4A - -
      $73  $4B - Left Arrow
      $8F  $4C - 5
      $74  $4D - Right arrow
      $4E  $4E - +
      $75  $4F - End
      $91  $50 - Down arrow
      $76  $51 - PgDn
      $92  $52 - Ins
      $93  $53 - Del
  }
  CtrlKeypadKeys: array[$47..$53] of byte =
    ($77, $8D, $84, $8E, $73, $8F, $74, $4E, $75, $91, $76, $92, $93);

begin
  Debug.DebugLog('xpcrt', Format('RemapScanCode: %d %d %d', [ScanCode, CtrlKeyState, KeyCode]), DLTrace);
  AltKey := ((CtrlKeyState AND
            (RIGHT_ALT_PRESSED OR LEFT_ALT_PRESSED)) > 0);
  CtrlKey := ((CtrlKeyState AND
            (RIGHT_CTRL_PRESSED OR LEFT_CTRL_PRESSED)) > 0);
  ShiftKey := ((CtrlKeyState AND SHIFT_PRESSED) > 0);
  if AltKey then begin
    Case KeyCode of
      VK_NUMPAD0 ..
      VK_NUMPAD9    : begin
                       DoingNumChars := true;
                       DoingNumCode := Byte((DoingNumCode * 10) + (KeyCode - VK_NUMPAD0));
                      end;
    end; { case }


    case ScanCode of
    // Digits, -, =
    $02..$0D: inc(ScanCode, $76);
    // Function keys
    $3B..$44: inc(Scancode, $2D);
    $57..$58: inc(Scancode, $34);
    // Extended cursor block keys
    $47..$49, $4B, $4D, $4F..$53:
              inc(Scancode, $50);
    // Other keys
    $1C:      Scancode := $A6;   // Enter
    $35:      Scancode := $A4;   // / (keypad and normal!)
    end
   end
  else if CtrlKey then
    case Scancode of
    // Tab key
    $0F:      Scancode := $94;
    // Function keys
    $3B..$44: inc(Scancode, $23);
    $57..$58: inc(Scancode, $32);
    // Keypad keys
    $35:      Scancode := $95;   // \
    $37:      Scancode := $96;   // *
    $47..$53: Scancode := CtrlKeypadKeys[Scancode];
    end
  else if ShiftKey then
    case Scancode of
    // Function keys
    $3B..$44: inc(Scancode, $19);
    $57..$58: inc(Scancode, $30);
    end
  else
    case Scancode of
    // Function keys
    $57..$58: inc(Scancode, $2E); // F11 and F12
    28: ScanCode := 13;
  end;
  Debug.DebugLog('xpcrt', Format('Result: %d ', [ScanCode]), DLTrace);
  Result := ScanCode;
end;

function KeyPressed : boolean;
var
  nevents, nread: dword;
  buf : TINPUTRECORD;
  AltKey: Boolean;
  wch: WideChar;
begin
  KeyPressed := FALSE;
  if ScanCode <> #0 then
    KeyPressed := TRUE
  else
  begin
     GetNumberOfConsoleInputEvents(StdInputHandle,nevents);
    while nevents>0 do
    begin
      ReadConsoleInputA(StdInputHandle,buf,1,nread);
      if buf.EventType = 2 {MOUSE_EVENT} then
      begin
        Result := UpdateMouseStatus(buf.Event.MouseEvent,ScanCode,SpecialKey);
      end
      else
      begin
        if buf.EventType = KEY_EVENT then
        begin
          if buf.Event.KeyEvent.bKeyDown then
          begin
          { Alt key is VK_MENU }
          { Capslock key is VK_CAPITAL }
          AltKey := ((Buf.Event.KeyEvent.dwControlKeyState AND
                     (RIGHT_ALT_PRESSED OR LEFT_ALT_PRESSED)) > 0);
          ShiftKeyState := ((Buf.Event.KeyEvent.dwControlKeyState AND SHIFT_PRESSED) > 0);
          CtrlKeyState := ((Buf.Event.KeyEvent.dwControlKeyState AND (RIGHT_CTRL_PRESSED OR LEFT_CTRL_PRESSED)) > 0);

          if not(Buf.Event.KeyEvent.wVirtualKeyCode in [VK_SHIFT, VK_MENU, VK_CONTROL,
                                               VK_CAPITAL, VK_NUMLOCK,
                                               VK_SCROLL]) then
            begin
              keypressed:=true;
  with Buf.Event.KeyEvent do
    Debug.DebugLog('xpcrt', Format('KeyPressed: %d %d %d %d', [wVirtualKeyCode, wVirtualScanCode, Ord(AsciiChar), Ord(UnicodeChar)]), DLTrace);

                      if (ord(buf.Event.KeyEvent.AsciiChar) = 0) or
                  (buf.Event.KeyEvent.dwControlKeyState and (LEFT_ALT_PRESSED or ENHANCED_KEY) > 0) then
              begin
                if Buf.Event.KeyEvent.wVirtualScanCode = $1C then // Num-Block-Enter
                  ScanCode := #13 else
                if (Buf.Event.KeyEvent.wVirtualScanCode = 16) and AltKey and CtrlKeyState then // Ctrl+Alt+Q in @ umwandeln
                  ScanCode := '@'
                else
                begin
                 SpecialKey := TRUE;
                 ScanCode := Chr(RemapScanCode(Buf.Event.KeyEvent.wVirtualScanCode, Buf.Event.KeyEvent.dwControlKeyState,
                                 Buf.Event.KeyEvent.wVirtualKeyCode));
                end;
              end
              else
              begin
                SpecialKey := FALSE;
                ScanCode := Chr(Ord(buf.Event.KeyEvent.AsciiChar));
                if (ScanCode = '?') and (buf.Event.KeyEvent.wVirtualKeyCode = 69) then
                  ScanCode := Chr($EE); // ''-Zeichen als 'e'-Zeichen mappen
                if (Ord(ScanCode) >= $80) then
                begin // Eingegebenes Zeichen grundstzlich nach Codepage 437 konvertieren
                  MultiByteToWideChar(GetConsoleCP,0,@ScanCode,1,@wch,1);
                  WideCharToMultiByte(437,0,@wch,1,@ScanCode,1,nil,nil);
                end;
              end;

              if Buf.Event.KeyEvent.wVirtualKeyCode in [VK_NUMPAD0..VK_NUMPAD9] then
              begin
                if AltKey then
                begin
                  Keypressed := false;
                  Specialkey := false;
                  ScanCode := #0;
                end
                else
                  break;
              end;
            end
          end
          else if (Buf.Event.KeyEvent.wVirtualKeyCode in [VK_MENU]) then
          begin
            if DoingNumChars then
            begin
              if DoingNumCode > 0 then
              begin
                ScanCode := Chr(DoingNumCode);
                Keypressed := true;

                DoingNumChars := false;
                DoingNumCode := 0;
                break
              end;
            end;
          end;
        end;
      end;
      { if we got a key then we can exit }
      if Result then
        exit;
      GetNumberOfConsoleInputEvents(StdInputHandle,nevents);
    end;
  end;
end;


function ReadKey: char;
begin
  while not KeyPressed do
    WaitForMultipleObjects(1,@StdInputHandle,true,INFINITE);

  if SpecialKey then begin
    Result := #0;
    SpecialKey := FALSE;
  end
  else begin
    Result := ScanCode;
    ScanCode := #0;
  end;

  Debug.DebugLog('xpcrt', Format('ReadKey: %d', [Integer(Result)]), DLTrace);
end;

procedure do_initialization;
var mode:DWORD;
begin
  StdInputHandle := GetStdHandle(STD_INPUT_HANDLE);

  mouse.maus:=GetSystemMetrics(SM_MOUSEPRESENT)<>0;

  if mouse.maus then
    if GetConsoleMode(StdInputHandle,mode) then
      if (mode and ENABLE_MOUSE_INPUT)=0 then
        SetConsoleMode(
          StdInputHandle,
          mode and not (ENABLE_ECHO_INPUT or ENABLE_LINE_INPUT or ENABLE_PROCESSED_INPUT) or ENABLE_MOUSE_INPUT
        );
end;

initialization
  do_initialization;

end.
